import numpy as np
from PIL import Image, ExifTags
from skimage.measure import block_reduce


class GraphTools(object):

    def __init__(self, serverless=False, debug=False):
        self.debug = debug
        self.serverless = serverless

    def calculate_kernel_size(self, image):
        # Determine kernel size from image
        image_h, image_w, *_ = image.shape

        kernel_height, image = self.find_lowest_denominator(image, image_length=image_h, edge='height')
        kernel_width, image = self.find_lowest_denominator(image, image_length=image_w, edge='width')

        kernel_size = (kernel_height, kernel_width)
        return image, kernel_size
    
    def crop_image(self, image, edge='both'):
        # Cut off a single pixel layer from the image
        if edge == 'height':
            return image[:image.shape[0]-1, :, :]
        elif edge == 'width':
            return image[:, :image.shape[1]-1, :]

        # Cut both edges
        return image[:image.shape[0]-1, :image.shape[1]-1, :]
    
    def edge_pixel_positions(self, edge_image):
        """
        Determine coordinates of edge pixels.
        :return: (N x 2) numpy array, where N is the number of edge pixels.
        """
        return np.argwhere(edge_image > 0.)
    
    def find_lowest_denominator(self, image, image_length=None, edge='height'):
        for factor in range(2, image_length // 2):
            if not image_length % factor:
                # Found the smallest denominator (stored in k_h)
                return factor, image
        
        # Remove one pixel layer off the image edge
        image = self.crop_image(image, edge=edge)
        return 2, image

    def image_mean(self, image_shape):
        # Determine mean image size
        return (image_shape[0] + image_shape[1]) / 2
    
    def reduce_image(self, image=None):
        # Calculate the smallest kernel size that fits into the image
        image, kernel = self.calculate_kernel_size(image)                

        # Reduce image size, given calculated kernel (max pooling)
        image = block_reduce(image, (kernel[0], kernel[1], 1), np.max)

        return image, kernel

    def save_image(self, image, filepath=None, split_rgb_channels=False):
        if split_rgb_channels:
            image = self.split_rgb_channels(image)

        if self.serverless:
            # Will need to set up to hit S3 here
            return
        
        # Otherwise save image in local results directory
        image_to_save = Image.fromarray(image)

        # See https://stackoverflow.com/questions/16720682/pil-cannot-write-mode-f-to-jpeg
        if image_to_save.mode != 'RGB':
            image_to_save = image_to_save.convert('RGB')

        image_to_save.save(filepath)

    @staticmethod
    def split_rgb_channels(image):
        return np.rot90(
            np.concatenate(
                (
                    np.concatenate(
                        (
                            np.rot90(image[:, :, 0], -1),
                            np.rot90(image[:, :, 1], -1),
                        )
                    ),
                    np.rot90(image[:, :, 2], -1),
                )
            )
        )
    
    @staticmethod
    def upright_image(image_filepath=None):
        '''
        Check for image orientation in exif data. See reference
        https://stackoverflow.com/questions/4228530/pil-thumbnail-is-rotating-my-image
        '''
        image = Image.open(image_filepath)
        if image._getexif() is not None:
            exif = dict(
                (ExifTags.TAGS[k], v)
                for k, v in image._getexif().items()
                if k in ExifTags.TAGS
            )
            if exif['Orientation'] == 3:
                return np.rot90(np.rot90(image))

            if exif['Orientation'] == 6:
                return np.rot90(image)

            if exif['Orientation'] == 8:
                return np.rot90(image, k=-1)

        return image



import os
import numpy as np
from skimage.measure import block_reduce

from clear_cut.utils.graph_tools import GraphTools


class BaseTracer(object):

    _graph_tools = None

    def __init__(self, method='Gradient', results_path=None, debug=False, serverless=True):
        self.debug = debug
        self.method = method
        self.serverless = serverless
        self.results_path = results_path or 'clear_cut/results/misc'

        if not self.serverless:
            self._get_or_create_results_dir()
    
    @property
    def graph_tools(self):
        if not self._graph_tools:
            if self.serverless:
                self._graph_tools = GraphTools(serverless=True, debug=self.debug)
                return self._graph_tools

            self._graph_tools = GraphTools(debug=self.debug)
        
        return self._graph_tools

    def merge_channels_of_traced_image(self, grdImg, origShape):
        """
        Merge gradImage RGB channels to one image
        """
        # Make image of correct shape
        xDim, yDim, chnls = origShape

        # Create empty array on the size of a single channel gradImage
        mrgdImg = np.zeros(shape=(2 * (xDim - 1), 2 * (yDim - 1)))

        # loop over each dimension, populating the gradient image
        x_offset = 2 * (yDim - 1)
        for i in range(0, 2 * (xDim - 1)):
            for j in range(0, 2 * (yDim - 1)):
                mrgdImg[i,j] = (
                    grdImg[i, j]
                    + grdImg[i, j + x_offset]
                    + grdImg[i, j + 2 * x_offset]
                )

        # Reduce gradient array to original image shape. Max pool gradient array using 2x2 kernel
        return block_reduce(mrgdImg, (2, 2), np.max)
        
    def _get_or_create_results_dir(self):
        results_path = self.results_path

        # Create results directory if it doesn't yet exist
        if '.' in results_path:
            results_path, _ = results_path.split('.')

        if not os.path.isdir(results_path):
            os.makedirs(results_path)



import os
import sys
import csv
import math
from random import randint

from clear_cut.utils.graph_tools import GraphTools


class ImageUtils(object):

    debug = False
    serverless = True
    _graph_tools = None

    def __init__(self):
        # increase csv file size limit
        csv.field_size_limit(sys.maxsize)

    @property
    def graph_tools(self):
        if not self._graph_tools:
            if self.serverless:
                self._graph_tools = GraphTools(serverless=True, debug=self.debug)
                return self._graph_tools

            self._graph_tools = GraphTools(debug=self.debug)
        
        return self._graph_tools

    def reduce_iter(self, i):
        return i - (i > 5) * 8

    def edge_kill(self, edg_img, coord, radius, wipe=False):
        """
        Determine whether the "shells border" are all non-edge pixels or not
        """
        # This may have already been wiped
        if not edg_img[coord[0], coord[1]]:
            if wipe:
                return edg_img

            return True

        # Initial counter and pre-define useful values
        border_size = 2 * radius + 1

        # Run over the square of pixels surrounding "radius" pixels around coord
        for i in range(0, border_size**2):
            dx = (i % border_size) - radius
            dy = (i // border_size) - radius

            x = coord[0] + dx
            y = coord[1] + dy

            if wipe:
                # Wipe whole shell of edge pixels
                try:
                    edg_img[x, y] = 0
                except IndexError:
                    # The central edge pixel is too close to the image perimeter, so ignore it
                    pass

                continue

            # Skip the inner shell pixels, just run around the border
            if abs(dx) != radius and abs(dy) != radius:
                continue
            
            try:
                if edg_img[x, y] > 0.:
                    # Found an edge pixel on the border, thus we cannot wipe this shell
                    return False
                
            except IndexError:
                # The central edge pixel is too close to the image perimeter, so ignore it
                continue

        if wipe:
            return edg_img

        # If we got here, all border pixels must be non-edge pixels
        return True

    def edge_killer(self, edge_image, pixel_tolerance=1):
        """
        Check and wipe out any edge pixels found within a pixel_tolerance radius.
        We refer to the radius of surround pixels as "the shell".
        """
        edge_coordinates = self.graph_tools.edge_pixel_positions(edge_image)

        for edge_coordinate in edge_coordinates:
            # Iterate over each layer of the shell (r --> r - 1 decrements)
            for sub_radius in reversed(range(1, pixel_tolerance + 1)):
                noisy_shell_found = self.edge_kill(edge_image, edge_coordinate, radius=sub_radius)

                if noisy_shell_found:
                    edge_image = self.edge_kill(edge_image, edge_coordinate, radius=sub_radius-1, wipe=True)
                    break

        return edge_image

    def __within_radius(self, x, y, chosen_one, R=10):
        return math.sqrt((x - chosen_one[0]) ** 2 + (y - chosen_one[1]) ** 2) <= R



import pydoc
import json
import time
import numpy as np
from collections import defaultdict
from random import randint

from clear_cut.utils.edge_utility import ImageUtils


class ClearCut(ImageUtils):

    _tracer = None

    def __init__(self, base_path='', debug=False, serverless=True, **kwargs):
        """
        If serverless, we must store results in S3 buckets
        """
        self.base_path = '/opt/python/lib/python3.7/site-packages/' if serverless else base_path
        self.debug = debug
        self.serverless = serverless
        
        self.base_dir = f'{self.base_path}clear_cut/images'
        self.default_image_selection(**kwargs)
    
    @property
    def tracer(self, method='gradient'):
        if not self._tracer:
            Tracer = pydoc.locate(
                f'clear_cut.utils.tracers.{method}.{str.capitalize(method)}Tracer'
            )
            self._tracer = Tracer(
                results_path=self.results_filepath,
                debug=self.debug,
                serverless=self.serverless,
            )
        
        return self._tracer

    def default_image_selection(self, **kwargs):
        self.image_filename = kwargs.get('image_filename', 'Bob.jpeg')

        self.image_filepath = '/'.join([self.base_dir, self.image_filename])
        self.image_size_threshold = kwargs.get('image_size_threshold', 600)
        self.pixel_tolerance = kwargs.get('pixel_tolerance', 10)

        self.image_raw = self.graph_tools.upright_image(image_filepath=self.image_filepath)
        self.image = np.array(self.image_raw)

        filename, _ = self.image_filename.split('.')
        self.results_filepath = f'{self.base_path}results/{filename}'
        self.reduce_image_size()

    def run(self):
        # Determine segmentation edges of the image (default method = gradient)
        edgy_image = self.tracer.trace_objects_in_image(image=self.image)

        # Reduce noise (edge pixels that cannot possibly contain an edge)
        edgy_image = self.edge_killer(edgy_image, pixel_tolerance=self.pixel_tolerance)

        self.graph_tools.save_image(
            edgy_image,
            filepath=f'{self.tracer.results_path}/0007_noise_reduced_image.png',
        )

        # Mask over the original image
        wipe_mask = edgy_image < 0.01
        bold_mask = edgy_image > 0.01
        self.image[wipe_mask] = 255
        self.image[bold_mask] = 0
        self.graph_tools.save_image(
            self.image,
            filepath=f'{self.tracer.results_path}/0008_edge_masked_image.png',
        )

    def reduce_image_size(self):
        # Build pooling dictionary
        pooling_history = defaultdict(lambda: defaultdict(tuple))
        pooling_history['iteration:0']['image_shape'] = self.image.shape

        # Check if the image is too small to be pooled, then pool the image
        while self.graph_tools.image_mean(self.image.shape) > self.image_size_threshold:
            image, kernel = self.graph_tools.reduce_image(image=self.image)
            
            # Update dictionary
            iter_no = 'iteration:{}'.format(len(pooling_history.keys()))
            pooling_history[iter_no] = {
                'image_shape': image.shape,
                'kernal_size': kernel,
            }
            
            # Must assign within the loop to dynamicaly update the while condition
            self.image = image
        
        # note that the final k is stored in "k"
        if self.debug:
            print('pooling_history={}'.format(
                json.dumps(pooling_history, indent=4)
            ))

            self.graph_tools.save_image(
                self.image,
                filepath='{}/0001_size_reduced_image.png'.format(self.tracer.results_path),
            )

            self.graph_tools.save_image(
                self.image,
                filepath='{}/0002_size_reduced_image_channel_collage.png'.format(self.tracer.results_path),
                split_rgb_channels=True,
            )



import os

from clear_cut.utils.tracers.base import BaseTracer


class TextureTracer(BaseTracer):

    def trace_objects_in_image(self, image=None, results_path=None, model_no=None):
        
        # object tracing texture method
        # create results/texture directory if it doesn't yet exist
        if not os.path.isdir(results_path):
            os.mkdir(results_path)

        timeIt = TimeCluster
        # gradImage: create numpy 2D array of size (2n-1) of the original
        dimY, dimX, chanls = origImage.shape

        # append an image (in x-direction) for each of the separate channels
        textureImage = np.zeros(shape=(dimY, dimX, 2), dtype = float)
        #textureImage = np.zeros(shape=(dimY, dimX, 3), dtype=float)

        # plot the red pixel pixel value versus the (r-g) % difference and (r-b) % difference
        # plot3D pre-setup
        #fig = plt.figure()
        # ax = fig.add_subplot(111, projection='3d')

        if timeIt:
            t_prev = time.time()
        # loop over each dimension, populating the textureImage with various labels
        remaining_pxls = []
        for j in range(0, dimY):
            #print("j/dimY =",j,"/",dimY)
            for i in range(0, dimX):

                # get difference between red and an `other' colour channel
                pt = (origImage[j, i])
                pt = [int(pt[0]), int(pt[1]), int(pt[2])]
                textureImage[j, i][0] = (pt[0] - pt[1])
                textureImage[j, i][1] = (pt[0] - pt[2])
                #textureImage[j, i][2] = pt[0]

                #print("textureImage[",j,",",i,"]=",textureImage[j,i])
                # get a coordinate list of unclassified pixel coordinate
                remaining_pxls.append([(pt[0] - pt[1]), (pt[0] - pt[2])])

                # plot3D layer creation
                #new_pt = textureImage[j, i]
                #ax.scatter(new_pt[0], new_pt[1], new_pt[2])

        if timeIt:
            print("\t \t 1. Populate textureImage: ", time.time() - t_prev, "seconds")
            t_prev = time.time()

        # plot the (r-g) % difference and (r-b) % difference
        plt.figure()
        plt.scatter(textureImage[:,:,0], textureImage[:,:,1], s=1)
        plt.xlabel("r-g")
        plt.ylabel("r-b")
        plt.show()

        # create new model using clustering algorithm
        model = ""
        # use existing model
        model = "10004078"

        # check if model path exists or not, then act accordingly
        if not os.path.isdir(results_path + "/model_" + str(model)):
            # create results/texture/model_{timestamp} directory
            model_path = results_path + "/model_" + str(model_no)
            os.mkdir(model_path)

            # keep track off original remaining pixels
            orig_remaining_pxls = remaining_pxls.copy()

            # classify clustered regions
            #print("remaining_pxls=", remaining_pxls)
            cluster_list = {
                "label_0" : []
            }
            #rad = 6
            rad = 5
            lbl_no = 0
            clr_map = plt.get_cmap('tab10')
            print("No. of remaining pxls (start) = ", len(remaining_pxls))
            print("cluster_list (start) = ", cluster_list)
            # keep finding clusters until all pxls have been labelled
            while len(orig_remaining_pxls) > 0:
                alive_direction = []
                dead_direction = []
                lbl_no += 1
                print("Cluster [", lbl_no,"] determination in progress...")

                # check the points in cluster label 1 have been removed
                '''if lbl_no == 2:
                    plt.figure()
                    plt.scatter((np.array(orig_remaining_pxls).T)[0], (np.array(orig_remaining_pxls).T)[1], s=1)
                    plt.show()'''

                if timeIt:
                    t_prev = time.time()
                # setup graph
                fig, ax = plt.subplots(1)
                ax.set_aspect('equal')
                #ax.scatter(textureImage[:, :, 0], textureImage[:, :, 1], s=1)
                ax.scatter((np.array(orig_remaining_pxls).T)[0], (np.array(orig_remaining_pxls).T)[1], s=1)
                if timeIt:
                    print("\t \t 2. Configure graph: ", time.time() - t_prev, "seconds")
                    t_prev = time.time()

                # put a criteria here to set all_outer_dead = True if all outer bubbles are dead...
                # edgy_img = np.zeros((255, 255))
                '''brdr = 2 * rad
                edgy_img = np.zeros((255 + 2 * brdr, 255 + 2 * brdr))
                for u in range(0, edgy_img.shape[0]):
                    for v in range(0, edgy_img.shape[1]):
                        # make a border around the image so that it does not go out of bounds?
                        if u < brdr or v < brdr or u > 255 + brdr or v > 255 + brdr:
                            edgy_img[u, v] = -0.1'''

                # randomly select a pixel coordinate in the existing list
                chosen_one = remaining_pxls[randint(0, len(remaining_pxls)-1)]
                #chosen_one = [255 // 2, 255 // 2]   # test case 1
                #chosen_one = [245, 0]   # test case 2
                #chosen_one = [0, 245]   # test case 3
                #chosen_one = [0, 2]     # test case 4
                #chosen_one = [250, 250]     # test case 5
                #print("chosen_one=",chosen_one)

                # initial nucleation
                #remaining_pxls, cluster_list, alive, dead = clstr_nucleate(chosen_one, rad, lbl_no, orig_remaining_pxls,
                #                                                           cluster_list, border=brdr, iter_max=17, init=True)
                #cluster_list, alive, dead = clstr_nucleate(chosen_one, rad, lbl_no, orig_remaining_pxls,
                #                                                     cluster_list, border=brdr, iter_max=17, init=True)
                cluster_list, alive, dead = clstr_nucleate(chosen_one, rad, lbl_no, orig_remaining_pxls,
                                                        cluster_list, iter_max=17, init=True)
                if timeIt:
                    print("\t \t 3. Cluster nucleate...: ", time.time() - t_prev, "seconds")
                    t_prev = time.time()

                alive_direction += alive
                dead_direction += dead

                # populate dead_directions with value 1 in edgy_img
                #for edge in dead_direction:
                #    edgy_img[edge[0], edge[1]] = 1

                #print("\t alive_direction=", alive_direction)
                #print("\t dead_direction=", dead_direction)

                # reiterate nucleation until all outer bubbles are dead
                #all_outer_dead = False
                # initiate new alive directions with current list
                #while not all_outer_dead:
                # for any outermost bubbles that found new pixels, nucleate another bubble around them
                #for dirs in alive_direction:
                idx = 0
                #while idx < len(alive_direction):
                #print("alive_direction=", alive_direction)
                while idx < len(alive_direction):
                    dirs = alive_direction[idx]
                    #print(" >>> dirs=", dirs)
                    idx += 1
                    #print("\t Iteration: ", idx,"/",len(alive_direction))

                    # re-nucleates on alive circles that are 2*rad distance away from initial point
                    #if abs(dirs[0]-chosen_one[0])==2*rad or abs(dirs[1]-chosen_one[1])==2*rad:

                    # re-nucleates on alive circles that are 2*rad distance away from initial point
                    if not ((dirs[0] == chosen_one[0]) and (dirs[1] == chosen_one[1])):
                        if timeIt:
                            t_prev = time.time()
                        #print("\t I'm outer alive!: ",dirs, "--> (", abs(dirs[0]-chosen_one[0]),",",abs(dirs[1]-chosen_one[1]))
                        cluster_list, alive, dead = clstr_nucleate(dirs, rad, lbl_no, orig_remaining_pxls, cluster_list)
                        if timeIt:
                            print("\t \t 3. Cluster nucleate...: ", time.time() - t_prev, "seconds")
                            t_prev = time.time()

                        # remove entries in alive that are already in alive_direction
                        for item in alive_direction:
                            # delete elements that are in alive_direction
                            # only increment index if the element is not in alive_direction
                            '''for curr in alive:
                                if item[0]==curr[0] and item[1]==curr[1]:
                                    alive.remove(curr)'''
                            if item in alive:
                                alive.remove(item)
                        if timeIt:
                            print("\t \t 4. Remove alive_direction duplicates: ", time.time() - t_prev, "seconds")
                            t_prev = time.time()

                        # make sure alive directions are not overwritten by dead ones!
                        alive_direction += alive

                        # remove duplicates in alive_direction
                        #alive_direction = np.unique(alive_direction, axis=0).tolist()

                        # remove alive directions if found in dead
                        for al in alive_direction:
                            '''for dd in dead:
                                if al[0] == dd[0] and al[1] == dd[1]:
                                    dead.remove(dd)'''
                            if al in dead:
                                dead.remove(al)
                        dead_direction += dead
                        #print("\t alive_direction=", alive_direction)
                        #print("\t dead_direction=", dead_direction)
                        if timeIt:
                            print("\t \t 5. Remove dead directions in alive_direction: ", time.time() - t_prev, "seconds")
                            t_prev = time.time()

                        #print("dead_direction=",dead_direction)
                        #print("No. of dead directions =", len(dead_direction))

                # graphics
                for coor in alive_direction:
                    ax.add_patch(Circle((coor[0], coor[1]), rad, facecolor=(0, 0, 0, 0), edgecolor=clr_map(lbl_no)))
                #for coor in dead_direction:
                #    ax.add_patch(Circle((coor[0], coor[1]), rad, facecolor=(0, 0, 0, 0), edgecolor=clr_map(lbl_no)))

                # plot the (r-g) % difference and (r-b) % difference
                plt.xlabel("r-g")
                plt.ylabel("r-b")
                plt.savefig(model_path+"/cluster_"+str(lbl_no)+".png")
                #plt.show()

                # remove duplicates in alive_direction
                alive_direction = np.unique(alive_direction, axis=0).tolist()

                # get all pixels inside the cluster
                enc_list = []
                for alv in alive_direction:
                    enc_list += cluster_counter(alv, remaining_pxls, R=rad, return_count=True)
                #print("len(enc_list)=",len(enc_list))

                # update cluster_list (this label) with pxls that were just found
                cluster_list["label_" + str(lbl_no)] += enc_list

                # remove any coordinates in remaining_pxls if they are with the cluster (enc_list)
                #t0 = time.time()
                updated_remaining_pxls = []
                for pxl in remaining_pxls:
                    if pxl not in enc_list:
                        updated_remaining_pxls.append(pxl)

                #print("Took ",time.time()-t0," seconds to remove enclosed pxls from remaining_pxls list")

                # repeat cluster finding until remaining_pxls is empty
                orig_remaining_pxls = updated_remaining_pxls
                remaining_pxls = orig_remaining_pxls.copy()
                #orig_remaining_pxls = remaining_pxls.copy()

                #print("No. of remaining pxls (end) = ", len(orig_remaining_pxls))
                #print("cluster_list (end) = ", cluster_list)

            # output final cluster_list dictionary
            print("Labelled all data!")

            # output dictionary to results folder
            write_file = csv.writer(open(model_path + "/cluster_list.txt", 'w'))
            for key, val in cluster_list.items():
                write_file.writerow([key, val])
        # if model directory does exist
        else:
            # read in cluster_list
            with open(results_path + "/model_" + str(model) + "/cluster_list.txt") as dict_file:
                reader = csv.reader(dict_file)
                cluster_list = {rows[0]:(rows[1].split()) for rows in reader}
        print("cluster_list=",cluster_list)

        # plot the (r-g) % difference and (r-b) % difference
        plt.xlabel("r-g")
        plt.ylabel("r-b")
        plt.show()

        exit()


        # Too small (shapes distinct but too much noise): 0.02
        # Maybe right? 0.07 (Bob.jpeg)
        # Too large (shaped not distinct enough): 0.10
        imCut = 0.08
        #imCut = 0.06
        # display gradient image
        '''plt.figure()
        plt.imshow(np.absolute(gradImage.T), interpolation="nearest")
        plt.figure()
        plt.imshow(np.multiply((np.absolute(gradImage.T) < (1-imCut)*255),(np.absolute(gradImage.T) > imCut*255)))'''

        # merge channels
        mrgIm1 = self.merge_channels_of_traced_image(gradImage.T, origImage.shape)
        #plt.figure()
        #plt.imshow(mrgIm1)

        edge_array = self.merge_channels_of_traced_image(
            np.multiply((np.absolute(gradImage.T) < (1 - imCut) * 255), (np.absolute(gradImage.T) > imCut * 255)),
            origImage.shape)
        #plt.figure()
        #plt.imshow(mrgIm2)

        # append 0s (non-edge pixels) to any missing columns/rows
        x_miss = origImage.shape[0] - edge_array.shape[0]
        if x_miss == 0:
            print("Same number of rows. Good!")
        elif x_miss > 0:
            print("Lost rows in compressing gradient. It can happen! Attempting to automatically dealing with it.")
            edge_array = np.concatenate((edge_array, np.zeros((1, edge_array.shape[1]))), axis = 0)
        else:
            print("Gained rows in compressing gradient. Doesn't make sense!")
            exit()

        y_miss = origImage.shape[1] - edge_array.shape[1]
        if y_miss == 0:
            print("Same number of columns. Good!")
        elif y_miss > 0:
            print("Lost columns in compressing gradient. It can happen! Attempting to automatically dealing with it.")
            edge_array = np.concatenate((edge_array, np.zeros((edge_array.shape[0], 1))), axis=1)
        else:
            print("Gained columns in compressing gradient. Doesn't make sense!")
            exit()

        # return an array of 0s (non-edges) and 1s (edges), same shape as passed in image
        print("Is ",image.shape," = ",edge_array.shape,"?")
        return edge_array
            